use std::io;
use crate::packet::PACKET_HEADER_SIZE;
use crate::packet::raw_packet::RawPacket;
use crate::prelude::as_u32_be;

impl TryFrom<Vec<u8>> for RawPacket {
    type Error = io::Error;
    fn try_from(data: Vec<u8>) -> io::Result<Self> {
        if data.len() < PACKET_HEADER_SIZE {
            return Err(io::Error::new(io::ErrorKind::InvalidData, "Can't parse data into RawPacket"))
        }
        let data_length = u16::from_be_bytes([data[0x12], data[0x13]]) as usize;
        if data.len() < 0x14 + data_length + 4 {
            return Err(io::Error::new(io::ErrorKind::InvalidData, "Can't parse data into RawPacket"))
        }
        let checksum = as_u32_be(data[(data.len() - 5)..].as_ref());
        if checksum != crc32fast::hash(data[..(data.len() - 4)].as_ref()) {
            return Err(io::Error::new(io::ErrorKind::InvalidData, "Checksum does not match"))
        }
        Ok(RawPacket {
            version: data[0],
            dest: as_u32_be(data[0x01..0x05].as_ref()),
            src: as_u32_be(data[0x05..0x09].as_ref()),
            packet_id: as_u32_be(data[0x09..0x0D].as_ref()),
            reply_to: as_u32_be(data[0x0D..0x11].as_ref()),
            command: data[0x11],
            data_length,
            data: data[0x14..data_length].to_vec(),
            checksum,
        })
    }
}

impl Into<Vec<u8>> for RawPacket {
    fn into(self) -> Vec<u8> {
        let mut v = vec![self.version];
        v.extend(self.dest.to_be_bytes());
        v.extend(self.src.to_be_bytes());
        v.extend(self.packet_id.to_be_bytes());
        v.extend(self.reply_to.to_be_bytes());
        v.push(self.command);
        v.extend(self.data_length.to_be_bytes());
        v.extend(self.data);
        v.extend(self.checksum.to_be_bytes());
        v
    }
}
